# coding: utf-8

from django.db.models import Model, CharField, TextField, URLField
from django.utils.translation import ugettext_lazy as _
from .managers import TermManager, CACHE_KEYS
from HTMLParser import HTMLParser
from django.core.cache import cache
from django.core.urlresolvers import reverse


class Term(Model):
    name = CharField(_('name'), max_length=100, unique=True,
                     help_text=_(u'Variants of the name can be specified with '
                                 u'a “|” separator (e.g. “Name|name|names”).'))
    definition = TextField(_('definition'), blank=True,
                           help_text=_('Accepts HTML tags.'))
    url = CharField(_('link'), max_length=200, blank=True, null=True,
                    help_text=_('Address to which the term will redirect '
                                '(instead of redirecting to the definition).'))

    objects = TermManager()

    class Meta:
        verbose_name = _('term')
        verbose_name_plural = _('terms')
        ordering = ('name',)

    def __unicode__(self):
        return self.original_name

    def save(self, *args, **kwargs):
        HTMLParser.unescape.__func__(HTMLParser, self.name)
        cache.delete_many(CACHE_KEYS)
        super(Term, self).save(*args, **kwargs)

    def get_absolute_url(self):
        if self.url:
            return self.url
        return reverse('term', kwargs={'pk': self.pk})

    def name_variants(self, variant_slice=slice(0, None)):
        return self.name.split('|')[variant_slice]

    @property
    def original_name(self):
        return self.name_variants(0)



# coding: utf-8

from django.db.models import Manager
from django.core.cache import cache
from django.template.loader import render_to_string
import re


VARIANTS_DICT_CACHE_KEY = 'terms__variants_dict'
REPLACE_DICT_CACHE_KEY = 'terms__replace_dict'
REPLACE_REGEXP_CACHE_KEY = 'terms__replace_regexp'
CACHE_KEYS = (VARIANTS_DICT_CACHE_KEY, REPLACE_DICT_CACHE_KEY,
              REPLACE_REGEXP_CACHE_KEY)


class TermManager(Manager):
    def variants_dict(self):
        d = cache.get(VARIANTS_DICT_CACHE_KEY)
        if d is None:
            d = {}
            for term in self.get_query_set().iterator():
                name_variants = term.name_variants()
                for variant in name_variants:
                    d[variant] = name_variants
            cache.set(VARIANTS_DICT_CACHE_KEY, d)
        return d

    def replace_dict(self):
        d = cache.get(REPLACE_DICT_CACHE_KEY)
        if d is None:
            d = {}
            template = 'terms/term_replace.html'
            for term in self.get_query_set().iterator():
                url = term.get_absolute_url()
                name_variants = term.name.split('|')
                context = {'url': url,
                           'url_is_external': bool(term.url)}
                for name_variant in name_variants:
                    context['name_variant'] = name_variant
                    d[name_variant] = render_to_string(template, context)
            cache.set(REPLACE_DICT_CACHE_KEY, d)
        return d

    def replace_regexp(self):
        r = cache.get(REPLACE_REGEXP_CACHE_KEY)
        if r is None:
            replace_dict = self.replace_dict()
            r = re.compile('(?P<before>^|\W)(?P<name>%s)(?P<after>\W|$)'
                           % '|'.join(map(re.escape, replace_dict)))
            cache.set(REPLACE_REGEXP_CACHE_KEY, r)
        return r



import os.path
from unittest import TestCase
from terms.html import NeutralHTMLReconstructor

CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))


class ValidHTMLTestCase(TestCase):
    def test(self):
        '''
        After being reconstructed, valid_html should be exactly the same.
        And after being reconstructed, valid_html_with_extra_spaces should
        be exactly the same as valid_html (since extra whitespaces within tags
        are stripped).
        '''
        filename = 'valid_html.html'
        html = open(os.path.join(CURRENT_PATH, filename)).read()
        filename = 'valid_html_with_extra_spaces.html'
        html_w_extra_spaces = open(os.path.join(CURRENT_PATH, filename)).read()

        r = NeutralHTMLReconstructor()
        r.feed(html)
        self.assertEqual(html, r.out)

        r.reset()
        r.feed(html_w_extra_spaces)
        self.assertEqual(html, r.out)



# coding: utf-8

from django.contrib.admin import site
from .models import Term
from .forms import TermForm
from django.conf import settings
from django.contrib.admin import ModelAdmin
if 'reversion' in settings.INSTALLED_APPS:
    from reversion import VersionAdmin as ModelAdmin


class TermAdmin(ModelAdmin):
    model = Term
    form = TermForm
    list_display = ('name', 'url')
    search_fields = ('name', 'definition', 'url')


site.register(Term, TermAdmin)



# coding: utf-8

from django.template import Library
from django.template.defaultfilters import stringfilter
from ..html import TermsHTMLReconstructor

register = Library()


@register.filter
@stringfilter
def replace_terms(html):
    parser = TermsHTMLReconstructor()
    parser.feed(html)
    return parser.out



# coding: utf-8

from django.forms import ModelForm, TextInput, ValidationError
from .models import Term
from django.utils.translation import ugettext as _
from .settings import AVAILABLE_WIDGETS, TERMS_DEFINITION_WIDGET as WIDGET
from django.conf import settings

# If WIDGET == 'auto': get the best widget one can import.
# Otherwise: Get the specified widget.
if WIDGET in AVAILABLE_WIDGETS[:2]:
    from django.forms import Textarea  # 'basic'
if WIDGET == AVAILABLE_WIDGETS[2] or (WIDGET == AVAILABLE_WIDGETS[0]
                                     and 'tinymce' in settings.INSTALLED_APPS):
    from tinymce.widgets import TinyMCE as Textarea  # 'tinymce'
if WIDGET == AVAILABLE_WIDGETS[3] or (WIDGET == AVAILABLE_WIDGETS[0]
                                    and 'ckeditor' in settings.INSTALLED_APPS):
    from ckeditor.widgets import CKEditorWidget as Textarea  # 'ckeditor'


class TermForm(ModelForm):
    def clean(self):
        definition = self.cleaned_data.get('definition')
        url = self.cleaned_data.get('url')
        if not definition and not url:
            raise ValidationError(_(u'Fill either “Definition” or “Link”.'))
        return super(TermForm, self).clean()

    class Meta:
        model = Term
        widgets = {
            'name': TextInput(attrs={'size': 120}),
            'definition': Textarea(),
        }



import os.path
from unittest import TestCase
from terms.html import TermsHTMLReconstructor

CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))


class InvalidHTMLTestCase(TestCase):
    def test_start_tag(self):
        '''
        After being reconstructed, invalid missing start tags should be there.
        '''
        filename = 'valid.html'
        valid = open(os.path.join(CURRENT_PATH, filename)).read()
        filename = 'missing_start_tag.html'
        invalid = open(os.path.join(CURRENT_PATH, filename)).read()

        r = TermsHTMLReconstructor()
        r.feed(invalid)
        self.assertEqual(valid, r.out)

    def test_end_tag(self):
        '''
        After being reconstructed, invalid missing end tags should be there.
        '''
        filename = 'valid.html'
        valid = open(os.path.join(CURRENT_PATH, filename)).read()
        filename = 'missing_end_tag.html'
        invalid = open(os.path.join(CURRENT_PATH, filename)).read()

        r = TermsHTMLReconstructor()
        r.feed(invalid)
        self.assertEqual(valid, r.out)

    def test_start_end_tag(self):
        '''
        After being reconstructed, invalid start-end tags should be valid.
        '''
        filename = 'valid.html'
        valid = open(os.path.join(CURRENT_PATH, filename)).read()
        filename = 'invalid_start-end_tag.html'
        invalid = open(os.path.join(CURRENT_PATH, filename)).read()

        r = TermsHTMLReconstructor()
        r.feed(invalid)
        self.assertEqual(valid, r.out)



from cms.menu_bases import CMSAttachMenu
from menus.base import NavigationNode
from menus.menu_pool import menu_pool
from django.core.urlresolvers import NoReverseMatch
from .models import Term
from django.utils.translation import ugettext_lazy as _


class TermMenu(CMSAttachMenu):
    name = _('Term Menu')

    def get_nodes(self, request):
        '''
        This method is used to build the menu tree.
        '''
        nodes = []
        for term in Term.objects.iterator():
            try:
                node = NavigationNode(
                    unicode(term),
                    term.get_absolute_url(),
                    term.pk,
                )
                nodes.append(node)
            except NoReverseMatch:
                pass
        return nodes


menu_pool.register_menu(TermMenu)



# coding: utf-8

from HTMLParser import HTMLParser
from .models import Term
from .settings import TERMS_IGNORED_TAGS, TERMS_IGNORED_CLASSES, \
                      TERMS_IGNORED_IDS, TERMS_REPLACE_FIRST_ONLY, TERMS_DEBUG
from .exceptions import HTMLValidationWarning


def concat_attrs(attrs):
        return ''.join(' %s="%s"' % attr for attr in attrs)


class NeutralHTMLReconstructor(HTMLParser):
    def reset(self):
        HTMLParser.reset(self)
        self.out = []
        self.out__append = self.out.append

    def feed(self, data):
        data = self.unescape(data)
        HTMLParser.feed(self, data)
        self.out = ''.join(self.out)

    def handle_startendtag(self, tag, attrs):
        self.out__append('<' + tag + concat_attrs(attrs) + ' />')

    def handle_starttag(self, tag, attrs):
        self.out__append('<' + tag + concat_attrs(attrs) + '>')

    def handle_endtag(self, tag):
        self.out__append('</' + tag + '>')

    def handle_data(self, data):
        self.out__append(data)

    def handle_comment(self, data):
        self.out__append('<!--' + data + '-->')

    def handle_decl(self, decl):
        self.out__append('<!' + decl + '>')

    def handle_pi(self, data):
        self.out__append('<?' + data + '>')

    def unknown_decl(self, decl):
        self.out__append('<![' + decl + ']>')


class TermsHTMLReconstructor(NeutralHTMLReconstructor):
    def reset(self):
        NeutralHTMLReconstructor.reset(self)
        self.tree_level = 0
        self.opened_tags = []
        self.opened_tags__append = self.opened_tags.append
        self.opened_tags__pop = self.opened_tags.pop
        self.disabled_level = None
        self.variants_dict = Term.objects.variants_dict()
        self.replace_dict = Term.objects.replace_dict()
        self.replace_regexp = Term.objects.replace_regexp()
        self.replace_regexp__sub = self.replace_regexp.sub

    @property
    def allow_replacements(self):
        return self.disabled_level is None

    def replace_terms(self, html):
        def translate(match):
            replace_dict = self.replace_dict
            match__group = match.group
            before, name, after = match__group('before'), \
                                  match__group('name'), \
                                  match__group('after')
            replaced_name = replace_dict.get(name, name)
            if TERMS_REPLACE_FIRST_ONLY and name in replace_dict:
                for variant in self.variants_dict[name]:
                    del replace_dict[variant]
            return before + replaced_name + after
        return self.replace_regexp__sub(translate, html)

    def handle_starttag(self, tag, attrs):
        NeutralHTMLReconstructor.handle_starttag(self, tag, attrs)
        self.opened_tags__append((tag, self.get_starttag_text(),
                                  self.getpos()))
        self.tree_level += 1

        dict_attrs = dict(attrs)
        dict_attrs__get = dict_attrs.get
        has_disabled_tag = tag in TERMS_IGNORED_TAGS
        classes = frozenset(dict_attrs__get('class', '').split())
        has_disabled_class = not classes.isdisjoint(TERMS_IGNORED_CLASSES)
        has_disabled_id = dict_attrs__get('id', '') in TERMS_IGNORED_IDS

        if self.allow_replacements and (has_disabled_tag or has_disabled_class
                                        or has_disabled_id):
            self.disabled_level = self.tree_level

    def handle_endtag(self, tag):
        try:
            opened_tag, full_start_tag, pos = self.opened_tags__pop()
            # Adds the tag to HTML only if it has a start tag.
            NeutralHTMLReconstructor.handle_endtag(self, tag)
        except IndexError:
            if TERMS_DEBUG:
                lines = self.rawdata.split('\n')
                line = self.getpos()[0]
                lines = '\n'.join(lines[line - 2:line])
                raise HTMLValidationWarning('unable to find the start tag '
                                            'for </%s> in:\n\n%s'
                                            % (tag, lines))
            return None
        if tag != opened_tag:
            if TERMS_DEBUG:
                lines = self.rawdata.split('\n')
                start_line, end_line = pos[0], self.getpos()[0]
                lines = '\n'.join(lines[start_line - 1:end_line])
                raise HTMLValidationWarning('unable to find the end tag '
                                            'for %s in:\n\n%s'
                                            % (full_start_tag, lines))
            self.tree_level -= 1  # We suppose the start tag is a start-end tag
                                  # with its final '/' missing.

        disabled_level = self.disabled_level
        if disabled_level and disabled_level == self.tree_level:
            self.disabled_level = None

        self.tree_level -= 1

    def handle_data(self, data):
        if self.allow_replacements:
            data = self.replace_terms(data)
        self.out__append(data)



#!/usr/bin/env python
# coding: utf-8

from setuptools import setup, find_packages
from terms import version_string


setup(
    name='django-terms',
    version=version_string,
    author='Bertrand Bordage',
    author_email='bordage.bertrand@gmail.com',
    url='https://github.com/BertrandBordage/django-terms',
    description='Site-wide adds a definition '
                'or a link for specialized terms.',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    license='BSD',
    packages=find_packages(),
    install_requires=[
        "Django >= 1.4",
    ],
    include_package_data=True,
    zip_safe=False,
)



from haystack.indexes import RealTimeSearchIndex, CharField, EdgeNgramField
from .models import Term


class TermIndex(RealTimeSearchIndex):
    text = CharField(document=True, use_template=True)
    content_auto = EdgeNgramField(model_attr='original_name')

    def get_model(self):
        return Term



from cms.app_base import CMSApp
from cms.apphook_pool import apphook_pool
from .menu import TermMenu
from django.utils.translation import ugettext_lazy as _


class TermApp(CMSApp):
    name = _('Term')
    urls = ['terms.urls']
    menus = [TermMenu]


apphook_pool.register(TermApp)
